@@ -106,7 +106,7 @@ class AsimovGenerator(AnalysisObject):
                                  poi_scale:Optional[float]=None):
         if asimov_types is None:
             asimov_types = self.DEFAULT_ASIMOV_TYPES
-        if isinstance(asimov_types, AsimovType):
+        if isinstance(asimov_types, (int, AsimovType)):
             asimov_types = [asimov_types]
         if asimov_names is not None:
             if isinstance(asimov_names, str):
