@@ -4,9 +4,10 @@ import os
 import re
 import ROOT
 
+import numpy as np
 import pandas as pd
 
-class RooProcAction(object):
+class RooProcBaseAction(object):
     def __init__(self, **params):
         self._params = params
         self.executed = False
@@ -41,17 +42,11 @@ class RooProcAction(object):
     
     def makedirs(self, fname):
         dirname = os.path.dirname(fname)
-        if not os.path.exists(dirname):
+        if dirname and (not os.path.exists(dirname)):
             os.makedirs(dirname)
-
-    def _execute(self, rdf, **kwargs):
-        return rdf
     
-    def execute(self, rdf, global_vars:Optional[Dict]=None):
-        params = self.get_formatted_parameters(global_vars)
-        rdf_next = self._execute(rdf, **params)
-        self.executed = True
-        return rdf_next
+    def execute(self, **params):
+        raise NotImplementedError
     
     @classmethod
     def parse_as_kwargs(cls, text):
@@ -69,8 +64,40 @@ class RooProcAction(object):
     @classmethod
     def parse(cls, text):
         raise NotImplementedError
+        
+class RooProcRDFAction(RooProcBaseAction):
+    def _execute(self, rdf, **params):
+        return rdf
     
-class RooProcTreeName(RooProcAction):
+    def execute(self, rdf, global_vars:Optional[Dict]=None):
+        params = self.get_formatted_parameters(global_vars)
+        rdf_next = self._execute(rdf, **params)
+        self.executed = True
+        return rdf_next         
+        
+class RooProcHelperAction(RooProcBaseAction):
+    def _execute(self, processor, **params):
+        return processor
+    
+    def execute(self, processor, global_vars:Optional[Dict]=None):
+        params = self.get_formatted_parameters(global_vars)
+        processor = self._execute(processor, **params)
+        self.executed = True
+        return processor
+
+class RooProcHybridAction(RooProcBaseAction):
+    
+    def _execute(self, rdf, processor, **params):
+        return rdf, processor
+    
+    def execute(self, rdf, processor, global_vars:Optional[Dict]=None):
+        params = self.get_formatted_parameters(global_vars)
+        rdf_next, processor_next = self._execute(rdf, processor, **params)
+        self.executed = True
+        return rdf_next, processor_next
+    
+    
+class RooProcTreeName(RooProcHelperAction):
     
     def __init__(self, treename:str):
         super().__init__(treename=treename)
@@ -79,7 +106,12 @@ class RooProcTreeName(RooProcAction):
     def parse(cls, text):
         return cls(treename=text)
     
-class RooProcExport(RooProcAction):
+    def _execute(self, processor, **params):
+        treename = params['treename']
+        processor.treename = treename
+        return processor
+    
+class RooProcExport(RooProcHelperAction):
     def __init__(self, filename:str):
         super().__init__(filename=filename)
         
@@ -88,7 +120,17 @@ class RooProcExport(RooProcAction):
         kwargs = cls.parse_as_kwargs(text)
         return cls(**kwargs)
     
-class RooProcGlobalVariables(RooProcAction):
+    def _execute(self, processor, **params):
+        filename = kwargs['filename']
+        data = {k:v.GetValue() for k,v in processor.external_variables.items()}
+        dirname = os.path.dirname(filename)
+        if dirname and (not os.path.exists(dirname)):
+            os.makedirs(dirname)
+        with open(filename, 'w') as outfile:
+            json.dump(data, outfile, indent=2)
+        return processor
+    
+class RooProcGlobalVariables(RooProcHelperAction):
     
     def __init__(self, **kwargs):
         super().__init__(**kwargs)
@@ -105,7 +147,11 @@ class RooProcGlobalVariables(RooProcAction):
             globs[result[1]] = result[2]
         return cls(**globs)
     
-class RooProcDefine(RooProcAction):
+    def _execute(self, processor, **params):
+        processor.global_variables.update(params)
+        return processor
+    
+class RooProcDefine(RooProcRDFAction):
     
     def __init__(self, name:str, expression:str):
         super().__init__(name=name, expression=expression)
@@ -119,13 +165,13 @@ class RooProcDefine(RooProcAction):
         expression = result.group(2)
         return cls(name=name, expression=expression)        
         
-    def _execute(self, rdf, **kwargs):
-        name = kwargs['name']
-        expression = kwargs['expression']
+    def _execute(self, rdf, **params):
+        name = params['name']
+        expression = params['expression']
         rdf_next = rdf.Define(name, expression)
         return rdf_next
     
-class RooProcFilter(RooProcAction):
+class RooProcFilter(RooProcRDFAction):
     def __init__(self, expression:str, name:Optional[str]=None):
         super().__init__(expression=expression,
                          name=name)
@@ -143,16 +189,16 @@ class RooProcFilter(RooProcAction):
             raise RuntimeError(f"multiple filter names detected in the expression `{text}`")
         return cls(name=name, expression=expression)
         
-    def _execute(self, rdf, **kwargs):
-        expression = kwargs['expression']
-        name = kwargs.get("name", None)
+    def _execute(self, rdf, **params):
+        expression = params['expression']
+        name = params.get("name", None)
         if name is not None:
             rdf_next = rdf.Filter(expression, name)
         else:
             rdf_next = rdf.Filter(expression)
         return rdf_next
     
-class RooProcSave(RooProcAction):
+class RooProcSave(RooProcHybridAction):
     
     def __init__(self, treename:str, filename:str, 
                  columns:Optional[List[str]]=None,
@@ -167,10 +213,10 @@ class RooProcSave(RooProcAction):
         kwargs = cls.parse_as_kwargs(text)
         return cls(**kwargs)
     
-    def _execute(self, rdf, **kwargs):
-        treename = kwargs['treename']
-        filename = kwargs['filename']
-        columns = kwargs.get('columns', None)
+    def _execute(self, rdf, processor, **params):
+        treename = params['treename']
+        filename = params['filename']
+        columns = params.get('columns', None)
         if columns is None:
             rdf_next = rdf.Snapshot(treename, filename)
         else:
@@ -181,9 +227,32 @@ class RooProcSave(RooProcAction):
             save_columns = list(set(save_columns))
             self.makedirs(filename)
             rdf_next = rdf.Snapshot(treename, filename, save_columns)
-        return rdf_next
+        processor.stdout.info(f"INFO: Writing output to `{filename}`.")
+        return rdf_next, processor
     
-class RooProcReport(RooProcAction):
+class RooProcAsNumpy(RooProcRDFAction):
+    
+    def __init__(self, filename:str, 
+                 columns:Optional[List[str]]=None):
+        super().__init__(filename=filename,
+                         columns=columns)
+        
+    @classmethod
+    def parse(cls, text):
+        kwargs = cls.parse_as_kwargs(text)
+        return cls(**kwargs)
+    
+    def _execute(self, rdf, **kwargs):
+        filename = kwargs['filename']
+        columns = kwargs.get('columns', None)
+        if columns is None:
+            data = rdf.AsNumpy()
+        else:
+            data = rdf.AsNumpy(columns)
+        np.save(filename, data)
+        return rdf    
+    
+class RooProcReport(RooProcHybridAction):
     def __init__(self, display:bool=False, filename:Optional[str]=None):
         super().__init__(display=display,
                          filename=filename)
@@ -192,9 +261,9 @@ class RooProcReport(RooProcAction):
         kwargs = cls.parse_as_kwargs(text)
         return cls(**kwargs)
     
-    def _execute(self, rdf, **kwargs):
-        display = kwargs['display']
-        filename = kwargs['filename']
+    def _execute(self, rdf, processor, **params):
+        display = params['display']
+        filename = params['filename']
         cut_report = rdf.Report()
         result = []
         cumulative_eff  = 1
@@ -209,17 +278,17 @@ class RooProcReport(RooProcAction):
             result.append(data)
         df = pd.DataFrame(result)
         if int(display):
-            print(df)
+            processor.stdout.info(df)
         if filename is not None:
             self.makedirs(filename)
             df.to_csv(filename)
-        return rdf
+        return rdf, processor
     
-class RooProcSum(RooProcAction):
+class RooProcSum(RooProcHybridAction):
     def __init__(self, ext_var_name:str, column_name:str):
         super().__init__(ext_var_name=ext_var_name,
                          column_name=column_name)
-        self.ext_var = {}
+        
     @classmethod
     def parse(cls, text):
         name_literals = re.findall(r"@{([^{}]+)}", text)
@@ -230,13 +299,13 @@ class RooProcSum(RooProcAction):
             raise RuntimeError("unspecified external variable name (format:@{ext_var_name})")
         return cls(ext_var_name=ext_var_name, column_name=column_name)
         
-    def _execute(self, rdf, **kwargs):
-        ext_var_name = kwargs['ext_var_name']
-        column_name = kwargs['column_name']
-        self.ext_var[ext_var_name] = rdf.Sum(column_name)
-        return rdf
+    def _execute(self, rdf, processor, **params):
+        ext_var_name = params['ext_var_name']
+        column_name = params['column_name']
+        processor.external_variables[ext_var_name] = rdf.Sum(column_name)
+        return rdf, processor
         
-class RooProcDeclare(RooProcAction):
+class RooProcDeclare(RooProcHelperAction):
     
     def __init__(self, expression:str, name:Optional[str]=None):
         super().__init__(expression=expression,
@@ -264,13 +333,13 @@ class RooProcDeclare(RooProcAction):
         status = ROOT.gInterpreter.Declare(declaration)
         return status
     
-    def _execute(self, rdf, **kwargs):
-        name = kwargs.get("name", None)
-        expression = kwargs['expression']
+    def _execute(self, processor, **params):
+        name = params.get("name", None)
+        expression = params['expression']
         self.declare_expression(expression, name)
-        return rdf
+        return processor
     
-class RooProcSaveFrame(RooProcAction):
+class RooProcSaveFrame(RooProcHelperAction):
     def __init__(self, name:str):
         super().__init__(name=name)
 
@@ -278,14 +347,26 @@ class RooProcSaveFrame(RooProcAction):
     def parse(cls, text):
         return cls(name=text)
     
-class RooProcLoadFrame(RooProcAction):
+    def _execute(self, processor, **params):
+        frame_name = params['name']
+        if frame_name in processor.rdf_frames:
+            processor.stdout.warning(f"WARNING: Overriding existing rdf frame `{frame_name}`")
+        processor.rdf_frames[frame_name] = processor.rdf
+        return processor
+    
+class RooProcLoadFrame(RooProcHelperAction):
     def __init__(self, name:str):
         super().__init__(name=name)
 
     @classmethod
     def parse(cls, text):
-        return cls(name=text)    
+        return cls(name=text)
     
+    def _execute(self, processor, **params):
+        frame_name = params['name']
+        if frame_name not in processor.rdf_frames:
+            raise RuntimeError(f"failed to load rdf frame `{frame_name}`: frame does not exist.")
+        processor.rdf = processor.rdf_frames[frame_name]
 
 ACTION_MAP = {
     'TREENAME': RooProcTreeName,
@@ -298,5 +379,6 @@ ACTION_MAP = {
     'GETSUM': RooProcSum,
     'EXPORT': RooProcExport,
     'SAVE_FRAME': RooProcSaveFrame,
-    'LOAD_FRAME': RooProcLoadFrame
+    'LOAD_FRAME': RooProcLoadFrame,
+    'AS_NUMPY': RooProcAsNumpy
 }
\ No newline at end of file
