@@ -66,39 +66,14 @@ class RooProcessor(AbstractObject):
         start = time.time()
         self.rdf = ROOT.RDataFrame(self.treename, all_files)
         for i, action in enumerate(self.action_list):
-            if isinstance(action, RooProcGlobalVariables):
-                self.global_variables.update(action._params)
-            elif isinstance(action, RooProcSaveFrame):
-                params = action.get_formatted_parameters(self.global_variables)
-                frame_name = params['name']
-                if frame_name in self.rdf_frames:
-                    self.stdout.warning(f"WARNING: Overriding existing rdf frame `{frame_name}`")
-                self.rdf_frames[frame_name] = self.rdf
-            elif isinstance(action, RooProcLoadFrame):
-                params = action.get_formatted_parameters(self.global_variables)
-                frame_name = params['name']
-                if frame_name not in self.rdf_frames:
-                    raise RuntimeError(f"failed to load rdf frame `{frame_name}`: frame does not exist.")
-                self.rdf = self.rdf_frames[frame_name]
-            elif isinstance(action, RooProcLoadFrame):
-                params = action.get_formatted_parameters(self.global_variables)
-            elif isinstance(action, RooProcExport):
-                data = {k:v.GetValue() for k,v in self.external_variables.items()}
-                params = action.get_formatted_parameters(self.global_variables)
-                filename = params['filename']
-                dirname = os.path.dirname(filename)
-                if not os.path.exists(dirname):
-                    os.makedirs(dirname)
-                with open(filename, 'w') as outfile:
-                    json.dump(data, outfile, indent=2)
-            else:
+            if isinstance(action, RooProcRDFAction):
                 self.rdf = action.execute(self.rdf, self.global_variables)
-                if isinstance(action, RooProcSave):
-                    params = action.get_formatted_parameters(self.global_variables)
-                    filename = params['filename']
-                    self.stdout.info(f"INFO: Writing output to `{filename}`.")
-            if isinstance(action, RooProcSum):
-                self.external_variables.update(action.ext_var)
+            elif isinstance(action, RooProcHelperAction):
+                action.execute(self, self.global_variables)
+            elif isinstance(action, RooProcHybridAction):
+                self.rdf, _ = action.execute(self.rdf, self, self.global_variables)
+            else:
+                raise RuntimeError("unknown action type")
         end = time.time()
         time_taken = end - start
         self.stdout.info(f"INFO: Task finished. Total time taken: {time_taken:.3f} s.")
\ No newline at end of file
