@@ -20,7 +20,7 @@ def release_tfiles():
     for f in ROOT.gROOT.GetListOfFiles():
         f.Close()
 
-def compile_macro(name='RooTwoSidedCBShape'):
+def compile_macro(name:str):
     macros_dir = os.path.join(quickstats.macro_path, 'macros')
     macros_path = os.path.join(macros_dir, name + '.cxx')
     print('INFO: Compiling macro "{}"...'.format(name))
@@ -29,7 +29,11 @@ def compile_macro(name='RooTwoSidedCBShape'):
         return -1
     return ROOT.gROOT.LoadMacro('{}++'.format(macros_path))
 
-def load_macro(name='RooTwoSidedCBShape'):  
+def load_macro(name:str):
+    auto_load = ROOT.gInterpreter.AutoLoad(name)
+    # already loaded by quickstats / third-party
+    if auto_load == 1:
+        return None
     macros_dir = os.path.join(quickstats.macro_path, 'macros')
     macros_path = os.path.join(macros_dir, name + '_cxx')
     if not os.path.exists(macros_path + '.so'):
